
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '32.8.3',
    'branch_nick': 'release-32-LTS',
    'revision_id': '51383ec0cd8981fe666f21149f0939f38cd0aa95',
    'revno': 12661
}
